<%@ WebHandler Language="C#" Class="CustomHandler.Writer" Debug="true" %>
using System;
using System.Web;
using System.Collections;
using System.Collections.Specialized;
using System.Text;

// 2004 Julien Couvreur (Dumky) : http://blog.monstuff.com
// This is distributed under the GPL : http://www.gnu.org/copyleft/gpl.html
// Based on Glen Murphy's dChat : http://sourceforge.net/projects/dchat/

// TODO:
// Add a message log (online and on file)
// Parse links in messages
// Add MSN Messenger integration (invite somebody)
// Add ink support
// Add .todo action
// Add .info action: lists the info for the current channel
// Improve focus behavior so that focus goes to input box more easily
// Add popup box when new message received, like Outlook Web Access

namespace CustomHandler
{

public class HandlerHelper {
    private HttpContext ctx = null;
    private string[] args = null;
    private string id = null;
    private Engine engine = null;
    private string username = null;
    public static bool DEBUG = true;
    
    public HandlerHelper (HttpContext oContext) {
        ctx = oContext;
        engine = Engine.GetInstance();
        Init();
    }
 
    public void Init() {
        string cmd = ctx.Request.QueryString["cmd"];
        Trace("cmd", cmd);

        id = ctx.Request.QueryString["id"];
        Trace("id", id);
        
        if (cmd != null) {
            // Parse command in the format of foo('test1','test2')

            cmd = cmd.Replace("(\'", "~");
            cmd = cmd.Replace("\',\'", "~");
            cmd = cmd.Replace("\')", "");
            cmd = cmd.Replace("()", "");
            
            string delimStr = "~";
            char [] delimiter = delimStr.ToCharArray();

            args = cmd.Split(delimiter);
            Trace("args", args);
        }
        
        username = GetUserName();
    }

    public void SetCacheHeaders() {
        ctx.Response.Cache.SetExpires(DateTime.Now.AddDays(-1));
        ctx.Response.Cache.SetCacheability(HttpCacheability.Public);
        ctx.Response.Cache.SetValidUntilExpires(false);

    }

    public void Exec() {
        // clean expired users
        StringCollection usersLeft = engine.ExpirePresence(10);
        foreach (string userLeft in usersLeft) {
            Trace("Exec. Expired user.", userLeft);
            //Output("chanmsg('" + userLeft + " has left the channel'); ");
            engine.AddMsg(new Message(userLeft, MsgType.Left));
        }
        
        switch (args[0]) {
            case "stream":
                bool bWelcome = false;
                bWelcome = !engine.UpdatePresence(username);
                if (bWelcome) {
                    engine.AddMsg(new Message(username, MsgType.Join));
                }
                ExecStream();
                if (bWelcome) {
                    Output("chanmsg('Welcome " + username + ". Type \\\".help\\\" for more info.'); ");		    
		        }
		        
                break;
            case "cmd":
                ExecCmd();
                break;
            case "users":
                ExecUsers();
                break;    
            default:
                Trace("Exec", "unrecognized");
                break;
        }

		if (id != null) {
		    Output("endFunc(" + id + ");");
        }
    }

    public string GetUserName() {
        if (DEBUG) {
            // load from the querystring
            string qs = ctx.Request.QueryString["user"];
            if (qs != null && qs.Length > 0) {
                return qs;
            }
        }
        
        // load from the windows creds
        string delimStr = "\\";
        char [] delimiter = delimStr.ToCharArray();

        string[] parseName = ctx.User.Identity.Name.Split(delimiter);
        if (parseName.Length > 1) {
            Trace("UserName", parseName[1]);
            return parseName[1];
        }
        
        // not ID found
        return "anonymous";
    }


    // list all messages in a range
    public void ExecStream() {
        int last = -1;
        if (args.Length >= 2) {
            try {
                last = Int32.Parse(args[1]);
            } catch (Exception e) {}
        } 
        Trace("ExecStream. Getting messages since.", last.ToString());
        ArrayList msgs = engine.GetMessagesSince(last);
        int total = last + msgs.Count;
        StringBuilder output = new StringBuilder();
        
        foreach (Message m in msgs) {
            switch (m.type) {
            case MsgType.Msg:
                // show all entries when you join, otherwise don't display my own entries
                if (last == -1 || m.user != username) {
                    // JCOUV TODO escape msg
                    output.Append("newmsg('" + m.user + "','" + m.msg + "'); ");
                }
                break;
            case MsgType.Join:
                output.Append("chanmsg('" + m.user + " has joined the channel'); ");
                break;
            case MsgType.Left:
                output.Append("chanmsg('" + m.user + " has left the channel'); ");
                break;
            case MsgType.Action:
                // JCOUV TODO find out what actnmsg are...
                //output.Append("actnmsg('$sysbody'); ");
                
                Trace("ExecStream. Found command.", m.msg);
                
                switch(GetAction(m.msg)) {
                case ActionType.Reset:
                    //output.Remove(0, output.Length);
                    output.Append("cleartxt(); chanmsg('" + m.user + " has reset the channel'); ");
                    break;
                }
  
                break;
            case MsgType.Start:
                output.Append("chanmsg('Server restarted'); ");
                break;
                
            }
        }
        
        Output(output.ToString());
        Output("laststream = " + total + "; "); 
    }
    
    
    // Execute a command (msg or action)
    public void ExecCmd() {
        if (args.Length < 2) {
            Trace("ExecCmd", "missing command");
            return;
        }
        string msg = args[1];
        
        msg = msg.Replace("<", "&lt;");
        msg = msg.Replace(">", "&gt;");
        msg = msg.Replace("\"", "&quot;");
        msg = msg.Replace("'", "&#39;");
        msg = msg.Replace("\\", "&#92;");

        if (msg.Length < 1) {
            Trace("ExecCmd", "empty command");
            return;
        }
        
        if (msg[0] == '.' && msg.Length > 2 && msg[1] != '.') {
            // handle actions (reset, help)
            msg = msg.Remove(0, 1);
            switch (GetAction(msg)) {
            case ActionType.Help:
                Output("chanmsg('******************* HELP ********************'); ");
                Output("chanmsg('Type \\\".help\\\" to get this message.      '); ");
                Output("chanmsg('Type \\\".reset\\\" to reset the channel.    '); ");
                Output("chanmsg('Type \\\".about\\\" for info on the software.'); ");
                Output("chanmsg('*********************************************'); ");
                
                break;
            case ActionType.Reset:    
                Trace("ExecCmd. Inserted reset command.", msg);
                int msgid = engine.AddMsg(new Message(username, msg, MsgType.Action));
                break;
            }
        } else {
            // insert message
            int msgid = engine.AddMsg(new Message(username, msg, MsgType.Msg));
            Trace("ExecCmd. Inserted message", msg);
            Trace("ExecCmd. ---- msgid", msgid.ToString());
        }        
    }
    
    // List the users that are present
    public void ExecUsers() {
        Trace("ExecUsers", "");
                
        Output("clearusers(); \n");
                
        foreach (string user in engine.ActiveUsersSince(60)) {
            Output("listuser('" + user + "'); \n");
        }
		Output("upduserslist(); \n");
    }
    
    
    public ActionType GetAction(String msg) {
        string delimStr = " ";
        string[] parts = null;
        char [] delimiter = delimStr.ToCharArray();

        parts = msg.Split(delimiter);
        if (parts.Length < 1) {
            return ActionType.None;
        }
        switch (parts[0]) {
        case "help":
            return ActionType.Help;
        case "reset":
            return ActionType.Reset;
        default:
            Trace("GetAction. No action matched", msg);
            return ActionType.None;
        }
    }
    
    public void Trace(string info, string s) {
        if (DEBUG) 
            ctx.Response.Write("//Trace[" + info + "]: " + (s==null?"null":s) + " (length=" + (s==null?"null":s.Length.ToString()) + ")<br>\n");
    }
    
    public void Trace(string info, string[] ss) {
        foreach (string s in ss) {
            Trace(info, s);
        }
    }
    
    public void Output(string s) {
        ctx.Response.Write(s);
    }
    
}

public class Engine {
    private static Engine instance = null; // singleton
    private Hashtable activeUsers = null;
    private ArrayList messages = null;
     
    public static Engine GetInstance() {
        // JCOUV TODO add thread safety
        if (instance == null) {
            instance = new Engine();
            instance.AddMsg(new Message(MsgType.Start));
        }
        return instance;
    } 
     
    private Engine() {
        activeUsers = new Hashtable();
        messages = new ArrayList();
    }
    
    ~Engine() {
        // save to file
        activeUsers = null;
        messages = null;
    } 
    
    public StringCollection ExpirePresence(int window) {
        lock(instance) {
            StringCollection ret = new StringCollection();
            
            foreach (string username in activeUsers.Keys) {
                DateTime lastseen = (DateTime) activeUsers[username];
                TimeSpan span = DateTime.Now - lastseen;
                if (span.TotalSeconds > window) {
                    ret.Add(username);
                }
            }
            
            foreach (string username in ret) {
                activeUsers.Remove(username);
            }
            return ret;
        }
    }
    
    // return false if this is an insert (not an update)
    public bool UpdatePresence(string username) {
        bool result = true;
        lock(instance) {
            result = (activeUsers[username] != null);
            activeUsers[username] = DateTime.Now;
        }
        return result;
    }
    
    public StringCollection ActiveUsersSince(int window) {
        lock(instance) {

            StringCollection ret = new StringCollection();
            
            foreach (string username in activeUsers.Keys) {
                DateTime lastseen = (DateTime) activeUsers[username];
                TimeSpan span = DateTime.Now - lastseen;
                if (span.TotalSeconds < window) {
                    ret.Add(username);
                }
            }
            return ret;
        }
    }
    
    // returns msgid
    public int AddMsg(Message msg) {
        lock(instance) {
            int count = messages.Count;
            messages.Add(msg);
            return count;
        }
    }
    
    // messages since msgid (excluded)
    public ArrayList GetMessagesSince(int msgid) {
        lock(instance) {
            if (messages.Count <= msgid) {
                return new ArrayList();
            }
            return messages.GetRange(msgid + 1, messages.Count - msgid - 1);
        }
    }
}

public enum MsgType { Msg, Start, Join, Left, Action }
public enum ActionType { None, Help, Reset }

public class Message {
    public string user;
    public string msg;
    public MsgType type;
    
    public Message(string _user, string _msg, MsgType _type) {
        user = _user;
        msg = _msg;
        type = _type;
    }
    
    public Message(string _user, MsgType _type) : this(_user, "", _type) { }

    public Message(MsgType _type) : this("", "", _type) { }

}


public class Writer : IHttpHandler
{ 
    public void ProcessRequest(HttpContext oContext)
    {
        HandlerHelper helper = new HandlerHelper(oContext);
        helper.SetCacheHeaders();
        helper.Exec();
    }

    public bool IsReusable { get { return true; } }
}
}
